/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// comndef.h  -- general definitions

#if !defined BYTE_DEFINED
typedef unsigned char 		byte;
#define BYTE_DEFINED 1
#endif

#undef true
#undef false

typedef enum {false, true} qboolean;

//============================================================================

typedef struct sizebuf_s
{
	qboolean	allowoverflow;	// if false, do a Sys_Error
	qboolean	overflowed;		// set to true if the buffer size failed
	byte		*data;
	int			maxsize;
	int			cursize;
	char		*function;
} sizebuf_t;

void SZ_Alloc (sizebuf_t *buf, int startsize);
void SZ_Clear (sizebuf_t *buf);
void *SZ_GetSpace (sizebuf_t *buf, int length);
void SZ_Write (sizebuf_t *buf, void *data, int length);
void SZ_Print (sizebuf_t *buf, char *data);	// strcats onto the sizebuf

//============================================================================

typedef struct link_s
{
	struct link_s	*prev, *next;
} link_t;

void ClearLink (link_t *l);
void RemoveLink (link_t *l);
void InsertLinkBefore (link_t *l, link_t *before);
void InsertLinkAfter (link_t *l, link_t *after);

#define	STRUCT_FROM_LINK(l,t,m) ((t *)((byte *)l - (int)&(((t *)0)->m)))

//============================================================================

#ifndef NULL
#define NULL ((void *)0)
#endif

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif
#ifndef max
#define max(a, b) ((a) > (b) ? (a) : (b))
#endif

#define bound(a, b, c) ((a) >= (c) ? (a) : (b) < (a) ? (a) : (b) > (c) ? (c) : (b))

#define Q_MAXCHAR ((char)0x7f)
#define Q_MAXSHORT ((short)0x7fff)
#define Q_MAXINT	((int)0x7fffffff)
#define Q_MAXLONG ((int)0x7fffffff)
#define Q_MAXFLOAT ((int)0x7fffffff)

#define Q_MINCHAR ((char)0x80)
#define Q_MINSHORT ((short)0x8000)
#define Q_MININT 	((int)0x80000000)
#define Q_MINLONG ((int)0x80000000)
#define Q_MINFLOAT ((int)0x7fffffff)

//============================================================================

extern	qboolean		bigendian;

extern	short	(*BigShort) (short l);
extern	short	(*LittleShort) (short l);
extern	int		(*BigLong) (int l);
extern	int		(*LittleLong) (int l);
extern	float	(*BigFloat) (float l);
extern	float	(*LittleFloat) (float l);

//============================================================================

void MSG_WriteChar (sizebuf_t *sb, int c);
void MSG_WriteByte (sizebuf_t *sb, int c);
void MSG_WriteShort (sizebuf_t *sb, int c);
void MSG_WriteByteShort (sizebuf_t *sb, int c);
void MSG_WriteByteShort2 (sizebuf_t *sb, int c, qboolean large);
void MSG_WriteLong (sizebuf_t *sb, int c);
void MSG_WriteFloat (sizebuf_t *sb, float f);
void MSG_WriteString (sizebuf_t *sb, char *s);
void MSG_WriteCoord (sizebuf_t *sb, float f);
void MSG_WriteAngle (sizebuf_t *sb, float f);

extern	int			msg_readcount;
extern	qboolean	msg_badread;		// set if a read goes beyond end of message

void	MSG_BeginReading (void);
int		MSG_ReadChar (void);
int		MSG_ReadByte (void);
int		MSG_ReadByteShort (void);
int		MSG_ReadByteShort2 (qboolean large);
int		MSG_SnoopByte (void);
int		MSG_ReadShort (void);
int		MSG_ReadLong (void);
float	MSG_ReadFloat (void);
char	*MSG_ReadString (void);

float	MSG_ReadCoord (void);
float	MSG_ReadAngle (void);

//============================================================================

#undef snprintf
#undef vsnprintf

#ifdef WIN32
# define snprintf _snprintf
# define vsnprintf _vsnprintf
#endif

int		va_snprintf (char *function, char *buffer, size_t buffersize, const char *format, ...);
int		va_vsnprintf (char *function, char *buffer, size_t buffersize, const char *format, va_list args);

#ifdef WIN32
#define strcasecmp _stricmp
#define strncasecmp _strnicmp
#endif

// strlcat and strlcpy, from OpenBSD
// Most (all?) BSDs already have them
#if defined(__OpenBSD__) || defined(__NetBSD__) || defined(__FreeBSD__) || defined(MACOSX)
# define HAVE_STRLCAT 1
# define HAVE_STRLCPY 1
#endif

#ifndef HAVE_STRLCAT
/*
* Appends src to string dst of size (unlike strncat, size is the
* full size of dst, no space left).  At most size - 1 characters
* will be copied.  Always NULL terminates (unless size <= strlen(dst)).
* Returns strlen(src) + MIN(size, strlen(initial dst)).
* If retval >= size, truncation occurred.
*/
size_t strlcat(char *dst, const char *src, size_t siz);
#define strlcat_s(dst,src)	strlcat(dst, src, sizeof(dst))
#endif  // #ifndef HAVE_STRLCAT

#ifndef HAVE_STRLCPY
/*
* Copy src to string dst of size.  At most size - 1 characters
* will be copied.  Always NULL terminates (unless size == 0).
* Returns strlen(src); if retval >= size, truncation occurred.
*/
size_t strlcpy(char *dst, const char *src, size_t siz);
#define strlcpy_s(dst,src)	strlcpy(dst, src, sizeof(dst))
#endif  // #ifndef HAVE_STRLCPY

//============================================================================

qboolean COM_Timeout (float *PrevTime, float WaitTime);

extern	char		com_token[1024];
extern	qboolean	com_eof;

char	*COM_ParseString (char *data);
int		COM_ParseToken (const char **datapointer, int returnnewline);

extern	int		com_argc;
extern	char	**com_argv;

int		COM_CheckParm (char *parm);
void	COM_Init (char *path);
void	COM_InitArgv (int argc, char **argv);
void	COM_CreatePath (char *path);
char	*COM_SkipPath (char *pathname);
void	COM_StripExtension (char *in, char *out);
void	COM_FileBase (const char *in, char *out);
void	COM_DefaultExtension (char *path, const char *extension, size_t len);

char *va(const char *format, ...);
// does a varargs printf into a temp buffer


//============================================================================

typedef int filehandle_t;

typedef enum {
	FS_READ,
	FS_WRITE,
	FS_APPEND
} packmode_t;

typedef enum {
	FS_SEEK_CUR,
	FS_SEEK_SET,
	FS_SEEK_END
} packorigin_t;

struct cache_user_s;
extern int		com_filesize;
extern  char	com_netpath[MAX_OSPATH];
extern	char	com_gamedir[MAX_OSPATH];
extern	char	com_basedir[MAX_OSPATH];

void	COM_InitFilesystem (void);
void	COM_ShutdownFileSystem (void);

void	COM_FSeek (filehandle_t f, int offset, packorigin_t origin);
int		COM_FTell (filehandle_t f);
int		COM_FGetc (filehandle_t f);
int		COM_FGetcLittleShort (filehandle_t f);
int		COM_FGetcLittleLong (filehandle_t f);

int		COM_SearchWildCards (char *searchstring, char **found, int maxlen, qboolean fullpath);

int		COM_FOpenDemoFile (char *filename, FILE **f);
int		COM_FOpenFile (char *name, filehandle_t *f, packmode_t mode);
int		COM_Read (void *buffer, int size, filehandle_t f);
int		COM_FRead (void *buffer, int size, int count, filehandle_t f);
void	COM_FCloseFile (filehandle_t f);

byte	*COM_LoadHeapFile (char *path);
byte	*COM_LoadTempFile (char *path);
byte	*COM_LoadHunkFile (char *path);

extern	struct cvar_s	registered;

extern qboolean		standard_quake, rogue, hipnotic;
